/* License
 * 
 * Copyright 1994-2004 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  
 *  * Redistribution of source code must retain the above copyright notice,
 *	this list of conditions and the following disclaimer.
 * 
 *  * Redistribution in binary form must reproduce the above copyright notice,
 *	this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL,
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE,
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *  
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility. 
 */

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.io.*;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

/**
* Connects to a telnet host using a user-specified host and port.
*/ 
public class MIDTerm extends MIDlet implements CommandListener, Runnable
{ 
    private Display display;
    private String host, port, user, pass;
    private TelnetCanvas canvas;
    private TelnetConnection connection;
    private InputStream input;
    private OutputStream output;
    private Command connectCommand, disconnectCommand, openCommand, closeCommand;
    private Command inputCommand, okCommand, cancelCommand, exitCommand;

    // form for login
    private Form loginForm;
    private TextField hostField;
    private TextField portField;
    
    // form for user input
    private Form inputForm;
    private TextField inputField;
    private ChoiceGroup inputOptions;
    private ChoiceGroup scrollOptions;
    
    // user-visible strings
    private final static String INPUT_ESC = "Prepend ESC";
    private final static String INPUT_CTRL = "Send as CTRL";
    private final static String INPUT_ENTER = "Append ENTER";
    private final static String INPUT_SCROLL = "Scroll Lock";
    
    public MIDTerm()
    {
        display = null;
    
        host = getAppProperty( "MIDTerm-Host" );
        if ( host == null ) host = "127.0.0.1";
        port = getAppProperty( "MIDTerm-Port" );
        if ( port == null ) port = "23";
        user = getAppProperty( "MIDTerm-User" );
        pass = getAppProperty( "MIDTerm-Pass" );
            
        canvas = new TelnetCanvas();
        connection = null;
        
        openCommand = new Command( "Open", Command.OK, 0 );
        exitCommand = new Command( "Exit", Command.EXIT, 0 );
        connectCommand = new Command( "Start", Command.OK, 0 );
        disconnectCommand = new Command( "Stop", Command.EXIT, 0 );
        closeCommand = new Command( "Close", Command.EXIT, 0 );
        inputCommand = new Command( "Input", Command.OK, 0 );
        okCommand = new Command( "OK", Command.OK, 0 );
        cancelCommand = new Command( "Cancel", Command.CANCEL, 0 );
        
        canvas.addCommand( closeCommand );
        canvas.addCommand( connectCommand );
        canvas.setCommandListener( this );
        
        // set up input form
        inputForm = new Form( "Input" );
        inputField = new TextField( null, "", 255, TextField.ANY );
        inputForm.append( inputField );
        inputOptions = new ChoiceGroup( null, Choice.MULTIPLE );
        inputOptions.append( INPUT_ENTER, null );
        inputOptions.append( INPUT_CTRL, null );
        inputOptions.append( INPUT_ESC, null );
        inputOptions.setSelectedIndex( 0, true ); // default ENTER to true
        inputForm.append( inputOptions );
        scrollOptions = new ChoiceGroup( null, Choice.MULTIPLE );
        scrollOptions.append( INPUT_SCROLL, null );
        inputForm.append( scrollOptions );
        inputForm.addCommand( okCommand );
        inputForm.addCommand( cancelCommand );
        inputForm.setCommandListener( this );
    }

    public void startApp()
    {
        if ( display == null )
        {
            display = Display.getDisplay( this );
            onShowLogin();
        }
    }

    public void pauseApp()
    {
        disconnect();
    }
    
    public void destroyApp( boolean unconditional )
    {
        disconnect();
    }
    
    private void connect()
    {
        disconnect();

        canvas.removeCommand( closeCommand );
        canvas.removeCommand( connectCommand );
        canvas.addCommand( disconnectCommand );
        canvas.addCommand( inputCommand );
        
        new Thread( this ).start();
    }
    
    private void disconnect()
    {
        canvas.removeCommand( disconnectCommand );
        canvas.removeCommand( inputCommand );
        canvas.addCommand( closeCommand );
        canvas.addCommand( connectCommand );
        
        if ( input != null )
        {
            try
            {
                input.close();
                input = null;
            }
            catch ( IOException ioe )
            {
                System.err.println( "Error while disconnecting input: " + ioe.toString() );
                canvas.receive( toASCII( "Error while disconnecting input: " + ioe.toString() ) );
            }
        }
        if ( output != null )
        {
            try
            {
                output.close();
                output = null;
            }
            catch ( IOException ioe )
            {
                System.err.println( "Error while disconnecting output: " + ioe.toString() );
                canvas.receive( toASCII( "Error while disconnecting output: " + ioe.toString() ) );
            }
        }
        if ( connection != null )
        {
            try
            {
                connection.close();
                connection = null;
            }
            catch ( IOException ioe )
            {
                System.err.println( "Error while disconnecting: " + ioe.toString() );
                canvas.receive( toASCII( "Error while disconnecting: " + ioe.toString() ) );
            }
        }
    }
    
    public void onShowLogin()
    {
        // create and populate login form
        loginForm = new Form( "Connect to" );
        hostField = new TextField( "Host", host, 50, TextField.URL );
        loginForm.append( hostField );
        portField = new TextField( "Port", port, 6, TextField.NUMERIC );
        loginForm.append( portField );
        loginForm.addCommand( exitCommand );
        loginForm.addCommand( openCommand );
        loginForm.setCommandListener( this );
        
        // show form
        display.setCurrent( loginForm );
    }

    public void onDoLogin()
    {
        host = hostField.getString().trim();
        port = portField.getString().trim();
        if ( host.length() == 0 ) host = "127.0.0.1";
        if ( port.length() == 0 ) port = "23";
        
        canvas.removeCommand( disconnectCommand );
        canvas.removeCommand( inputCommand );
        canvas.addCommand( closeCommand );
        canvas.addCommand( connectCommand );
        canvas.reset();
        
        display.setCurrent( canvas );
        
        connect();
        
        // dispose login form
        loginForm = null;
        hostField = null;
        portField = null;
    }
    
    public void onShowInput()
    {
        scrollOptions.setSelectedIndex( 0, canvas.isScrolling() );
        display.setCurrent( inputForm );
    }
    
    public void onDoInput()
    {
        canvas.setScrolling( scrollOptions.isSelected( 0 ) );

        StringBuffer buffer = new StringBuffer( inputField.getString() );
    
        if ( inputOptions.isSelected( 0 ) ) // ENTER
        {
            buffer.append( '\n' );
        }
        
        if ( inputOptions.isSelected( 2 ) ) // ESC
        {
            buffer.insert( 0, (char) 27 );
        }
        
        byte[] data = toASCII( buffer.toString() ); 

        if ( inputOptions.isSelected( 1 ) ) // CTRL
        {
            // convert any alpha characters to control characters
            byte b;
            for ( int i = 0; i < data.length; i++ )
            {
                b = data[i];
                if ( b > 96 && b < 123 ) b = (byte) (b - 32);
                if ( b > 63 && b <  96 ) b = (byte) (b - 64);
                data[i] = b;
            }
        }

        if ( data.length > 0 )
        {
            canvas.send( data );
        }
    }   
    
    public void onHideInput()
    {
        display.setCurrent( canvas );
        inputField.setString( "" ); // clear last command if any
        // inputOptions.setSelectedIndex( 0, false ); // ENTER // convenience
        inputOptions.setSelectedIndex( 1, false ); // CTRL
        inputOptions.setSelectedIndex( 2, false ); // ESC
    }
    
    // interface CommandListener
    
    public void commandAction( Command aCommand, Displayable aDisplayable )
    {
        if ( aCommand == openCommand )
        {
            onDoLogin();
        }
        else
        if ( aCommand == closeCommand )
        {
            onShowLogin();
        }
        else
        if ( aCommand == connectCommand )
        {
            connect();
        }
        else
        if ( aCommand == disconnectCommand )
        {
            disconnect();
        }
        else
        if ( aCommand == exitCommand )
        {
            destroyApp( true );
            notifyDestroyed();
        }
        else
        if ( aCommand == inputCommand )
        {
            onShowInput();
        }
        else
        if ( aCommand == okCommand )
        {
            onDoInput();
            onHideInput();
        }
        else
        if ( aCommand == cancelCommand )
        {
            onHideInput();
        }
    }
    
    // interface Runnable
    
    public void run()
    {
        String connectString = "socket://" + host + ':' + port;
        
        try
        {
            canvas.receive( toASCII( "Connecting...\n" ) );
            
            connection = new TelnetConnection( (StreamConnection) 
                Connector.open( connectString, Connector.READ_WRITE, true ),
                canvas.getColumns(), canvas.getRows(), canvas.getTerminalType() );
            input = connection.openInputStream();
            output = connection.openOutputStream();
            canvas.setOutputStream( output );
            
            try
            {        
                if ( user != null )
                {
                    waitUntil( input, new String[] { "ogin:", "serid:" }, true );
                    output.write( toASCII( user + '\n' ) );
                    canvas.receive( toASCII( user + '\n' ) );
                    output.flush();

                    if ( pass != null )
                    {
                        waitUntil( input, new String[] { "assword:" }, true );
                        output.write( toASCII( pass + '\n' ) );
                        canvas.receive( toASCII( "\n" ) ); // let's not echo password
                        output.flush();
                    }
                }

                waitUntil( input, new String[] { "Done" }, true );
            }
            catch ( IOException ioe )
            {
                System.err.println( "Error while communicating: " + ioe.toString() );
                canvas.receive( toASCII( "\nLost connection." ) );
            }
            catch ( Throwable t )
            {
                System.err.println( "Unexpected error while communicating: " + t.toString() );
                canvas.receive( toASCII( "\nUnexpected error: " + t.toString() ) );
            }
        }
        catch ( IllegalArgumentException iae )
        {
            System.err.println( "Invalid host: " + host );
            canvas.receive( toASCII( "Invalid host: " + host ) );
        }        
        catch ( ConnectionNotFoundException cnfe )
        {
            System.err.println( "Connection not found: " + connectString );
            canvas.receive( toASCII( "Connection not found: " + connectString ) );
        }        
        catch ( IOException ioe )
        {
            System.err.println( "Error on connect: " + ioe.toString() );
            canvas.receive( toASCII( "Error on connect: " + ioe.toString() ) );
        }        
        catch ( Throwable t )
        {
            System.err.println( "Unexpected error on connect: " + t.toString() );
            canvas.receive( toASCII( "Unexpected error on connect: " + t.toString() ) );
        }        
        
        disconnect();
        canvas.receive( toASCII( "\nDisconnected.\n" ) );
    }
    
    /**
    * Returns the String that matched, or null if end of input.
    */
    private String waitUntil( 
        InputStream inInput, String[] inMatches, boolean append )
    throws IOException
    {
        int c, i;
        
        // no input or zero matches returns immediately
        if ( inInput == null || inMatches == null || inMatches.length == 0 ) 
        {
            return null;
        }

        // zero-length strings return immediately
        for ( i = 0; i < inMatches.length; i++ )
        {
            if ( inMatches[i].length() == 0 ) return inMatches[i];
        }

        // create match trackers (init to zeroes)
        int[] lastMatch = new int[ inMatches.length ];
        
        // now read input and check for match
        while ( ( c = inInput.read() ) != -1 )
        {
            if ( append ) canvas.receive( (byte) c );
                        
            for ( i = 0; i < inMatches.length; i++ )
            {
                if ( inMatches[i].charAt( lastMatch[i] ) == c )
                {
                    // increment match count
                    if ( ++lastMatch[i] == inMatches[i].length() )
                    {
                        // matched complete string: return
                        return inMatches[i];
                    }
                }
                else // no match
                {
                    // reset match count
                    lastMatch[i] = 0;
                }
            }
        }
        
        // end of input
        return null;
    }
    
    private final static byte[] toASCII( String inString )
    {
        char[] chars = inString.toCharArray();
        byte[] bytes = new byte[ chars.length ];
        for ( int i = 0; i < bytes.length; i++ ) bytes[i] = (byte) chars[i];
        return bytes;
    }
}